HITS_INFO = {
    # TODO(ronak): Add cached results from retrieval methods
    # BM25
    # "BM25/retrieve_results_researchy_questions_test_top100.json": {
    #     "description": "Top-100 BM25 Results for a subset of the Researchy Questions task. (Cached JSON format)",
    #     "urls": [
    #         "https://github.com/castorini/rank_llm_data/raw/main/retrieve_results/BM25/retrieve_results_covid_top100.json"
    #     ],
    #     "md5": "84ee088e57c6bd8e274b25ce45406317",
    #     "size (bytes)": 9113367,
    #     "downloaded": False,
    # },
    # RANK_ZEPHYR
    # "RANK_ZEPHYR/retrieve_results_covid_top100.json": {
}
